package pl.mimuw.wfnetreduction.algorithms;

import static pl.mimuw.wfnetreduction.algorithms.PetrinetHelper.getInPlaces;
import static pl.mimuw.wfnetreduction.algorithms.PetrinetHelper.getInTransitions;
import static pl.mimuw.wfnetreduction.algorithms.PetrinetHelper.getOutPlaces;
import static pl.mimuw.wfnetreduction.algorithms.PetrinetHelper.getOutTransitions;

import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

import com.google.common.collect.Sets;

import pl.mimuw.wfnetreduction.parameters.WFNetReductionParameters;

public class AndOrReductionAlgorithm {

	LazyIterator<Pair<Place, Place>> iteratorPP;
	LazyIterator<Pair<Transition, Transition>> iteratorTT;
	ContractionLog log;

	public void init(Petrinet net_) {
		iteratorPP = LazyIterator.from(PetrinetHelper.iteratePairsOfDownstreamPlacesTransitiveClosure(net_));
		iteratorTT = LazyIterator.from(PetrinetHelper.iteratePairsOfDownstreamTransitionsTransitiveClosure(net_));
	}
	
	/**
	 * The method that implements your algorithm.
	 * 
	 * Note that this method only uses the boolean which is stored in the
	 * parameters. Nevertheless, it could have used the integer and/or the String as
	 * well.
	 * 
	 * @param context
	 *            The context where to run in.
	 * @param net
	 *            Petri net
	 * @param intputs
	 *            Set of input places
	 * @param outputs
	 *            Set of output places
	 * @param parameters
	 *            The parameters to use.
	 * @return The results of the reduction procedure which is a boolean that says
	 *         if the net is an AND-OR net.
	 */
	public ContractionLog apply(PluginContext context, Petrinet net, Set<PetrinetNode> inputs,
			Set<PetrinetNode> outputs, WFNetReductionParameters parameters) {
		if (inputs.isEmpty() || outputs.isEmpty())
			throw new IllegalArgumentException("[AND-OR reducer] Empty interface set");
		if (inputs.iterator().next().getClass() != outputs.iterator().next().getClass())
			throw new IllegalArgumentException("[AND-OR reducer] Input and output interfaces are of different type");
		/**
		 * Put your algorithm here, which computes an output form the inputs provided
		 * the parameters.
		 */
		long time = -System.currentTimeMillis();
		//System.err.println("System.err");
		parameters.displayMessage("[AND-OR reducer] Start");
		parameters.displayMessage("[AND-OR reducer] Net = " + net.toString());
		parameters.displayMessage("[AND-OR reducer] Inputs = " + inputs.toString());
		parameters.displayMessage("[AND-OR reducer] Outputs = " + outputs.toString());
		parameters.displayMessage("[AND-OR reducer] Limit to 11 = " + parameters.isLimitTo11());
		parameters.displayMessage("[AND-OR reducer] Parameters = " + parameters.toString());
		if (inputs == null || inputs.size() == 0) {
			throw new IllegalArgumentException(
					"[AND-OR reducer] Cannot proceeded with no inputs. Please mark every input place with 1 token or with 3 tokens if it is supposed to be input and output place at the same time.");
		}
		if (outputs == null || outputs.size() == 0) {
			throw new IllegalArgumentException(
					"[AND-OR reducer] Cannot proceeded with no outputs. Please mark every output place with 2 tokens or with 3 tokens if it is supposed to be input and output place at the same time.");
		}

		Petrinet net_ = null;
		if (net.getNodes().size() <= 1) {
			log = new ContractionLog(net);
			log.setResult(true);
			log.println("Single node net is hierarchical :)");
			log.nextStep();
			return log;
		} else {
			log = new ContractionLog(net);
			log.println("starting contractions");
			Triple<Petrinet, Set<PetrinetNode>, Set<PetrinetNode>> cloned = PetrinetHelper.clonePetrinet(net, inputs,
					outputs);
			net_ = cloned.getLeft();
			Set<PetrinetNode> inputs_ = cloned.getMiddle();
			Set<PetrinetNode> outputs_ = cloned.getRight();

			init(net_);
			
			boolean somethingChanged = true;
			whileEt: while (somethingChanged) {
				log.nextStep(net_);
				log.println("Current net: "+PetrinetHelper.toString(Triple.of(net_, inputs_, outputs_)));
				//System.err.println("Current net: "+PetrinetHelper.toString(Triple.of(net_, inputs_, outputs_)));

				somethingChanged = false;
				if (!parameters.isLimitTo11()) {
				    somethingChanged = joinMultiinterface(net_, inputs_, outputs_, log);
				    if (somethingChanged)
					    continue whileEt;
				}
				somethingChanged = mergeLoops(net_, inputs_, outputs_, log);
				if (somethingChanged)
					continue whileEt;
				somethingChanged = contractSomeExpandedSubnet(net_, inputs_, outputs_, log, parameters.isLimitTo11());
			}
		}

		time += System.currentTimeMillis();
		parameters.displayMessage("[AND-OR reducer] Output = " + (net_.getNodes().size() == 1));
		parameters.displayMessage("[AND-OR reducer] End (took " + time / 1000.0 + "  seconds).");
		log.println("Is the net hierarchical: " + log.getResult());
		log.nextStep();
		log.setResult(net_.getNodes().size() == 1);
		try {
		log.storeResultNetToFile("/home/jsroka/net.tpn");
		} catch (IOException e) {}
		return log;
	}

	int contractionCount = 0;

	private final String getContractedName(String prefix) {
		return prefix + contractionCount++;
	}

	private final void contractGivenPSubnet(Petrinet netN, Set<PetrinetNode> inputsN, Set<PetrinetNode> outputsN,
			Set<PetrinetNode> netM, Set<PetrinetNode> inputsM, Set<PetrinetNode> outputsM, String prefix, ContractionLog log) {
		Set<Transition> inputsForContracted = getInTransitions(netN, (Place) inputsM.iterator().next());
		inputsForContracted.removeAll(netM);
		Set<Transition> outputsForContracted = getOutTransitions(netN, (Place) outputsM.iterator().next());
		outputsForContracted.removeAll(netM);

		Place p = netN.addPlace(getContractedName(prefix));
		//System.err.println("contracting into P=" + p.getLabel() + " of " + netM + " I=" + inputsM + " O=" + outputsM);
		//System.err.println("inputs for contracted:" + inputsForContracted);
		//System.err.println("outputs for contracted:" + outputsForContracted);
		log.println("contracting into P=" + p.getLabel() + " of " + netM + " with I=" + inputsM + " O=" + outputsM);
		log.addContractedNodes(netM);

		for (PetrinetNode n : netM) {
			netN.removeNode(n);
		}

		if (inputsN.removeAll(inputsM))
			inputsN.add(p);
		if (outputsN.removeAll(outputsM))
			outputsN.add(p);

		for (Transition t : inputsForContracted) {
			//System.err.println("new place name: "+p);
			//System.err.println("nodes of the net after contraction: " + netN.getNodes().toString());
			netN.addArc(t, p);
		}
		for (Transition t : outputsForContracted) {
			//System.err.println("nodes of the net after contraction: " + netN.getNodes().toString());
			netN.addArc(p, t);
		}

		PetrinetHelper.filterDownstreamStreamsP(iteratorPP, iteratorTT, netM, inputsM, outputsM, p);
	}

	private final void contractGivenTSubnet(Petrinet netN, Set<PetrinetNode> inputsN, Set<PetrinetNode> outputsN,
			Set<PetrinetNode> netM, Set<PetrinetNode> inputsM, Set<PetrinetNode> outputsM, String prefix, ContractionLog log) {
		Set<Place> inputsForContracted = getInPlaces(netN, (Transition) inputsM.iterator().next());
		inputsForContracted.removeAll(netM);
		Set<Place> outputsForContracted = getOutPlaces(netN, (Transition) outputsM.iterator().next());
		outputsForContracted.removeAll(netM);

		Transition t = netN.addTransition(getContractedName(prefix));
		//System.err.println("contracting into T=" + t.getLabel() + " of " + netM + " I=" + inputsM + " O=" + outputsM);
		//System.err.println("inputs for contracted:" + inputsForContracted);
		//System.err.println("outputs for contracted:" + outputsForContracted);
		log.println("contracting into T=" + t.getLabel() + " of " + netM + " with I=" + inputsM + " O=" + outputsM);
		log.addContractedNodes(netM);

		for (PetrinetNode n : netM) {
			netN.removeNode(n);
		}

		if (inputsN.removeAll(inputsM))
			inputsN.add(t);
		if (outputsN.removeAll(outputsM))
			outputsN.add(t);

		for (Place p : inputsForContracted) {
			netN.addArc(p, t);
		}
		for (Place p : outputsForContracted) {
			netN.addArc(t, p);
		}

		PetrinetHelper.filterDownstreamStreamsT(iteratorPP, iteratorTT, netM, inputsM, outputsM, t);
	}

	private final boolean contractSomeExpandedSubnet(Petrinet net, Set<PetrinetNode> inputs,
			Set<PetrinetNode> outputs, ContractionLog log, boolean limitTo11) {
		//System.err.println("trying to expand pairs");
		log.println("trying to expand pairs");
		while (iteratorPP.hasNext()) {
			Pair<Place, Place> p = iteratorPP.next();
			//System.err.println("next P pair to expand:" + p);
			log.println("next P pair to expand:" + p);
			Triple<Set<PetrinetNode>, Set<PetrinetNode>, Set<PetrinetNode>> expanded = PetrinetHelper.expandP(net,
					inputs, outputs, p, log, limitTo11);
			if (expanded != null) {
				contractGivenPSubnet(net, inputs, outputs, expanded.getLeft(), expanded.getMiddle(),
						expanded.getRight(), "expandP", log);
				return true;
			}
		}

		while (iteratorTT.hasNext()) {
			Pair<Transition, Transition> t = iteratorTT.next();
			//System.err.println("next T pair:" + t);
			log.println("next T pair to expand:" + t);
			Triple<Set<PetrinetNode>, Set<PetrinetNode>, Set<PetrinetNode>> expanded = PetrinetHelper.expandT(net,
					inputs, outputs, t, log, limitTo11);
			if (expanded != null) {
				contractGivenTSubnet(net, inputs, outputs, expanded.getLeft(), expanded.getMiddle(),
						expanded.getRight(), "expandT", log);
				return true;
			}
		}

		return false;
	}

	public final boolean mergeLoops(Petrinet net, Set<PetrinetNode> inputs, Set<PetrinetNode> outputs, ContractionLog log) {
		//System.err.println("trying to merge loops");
		log.println("trying to merge loops");
		for (Place p : net.getPlaces()) {
			Set<Transition> transOut = getOutTransitions(net, p);
			Set<Transition> transOutIn = new HashSet<>();
			for (Transition t: transOut) {
				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = net.getOutEdges(t);
				if ((outEdges.size() == 1) && outEdges.stream().allMatch(edge -> edge.getTarget().equals(p)))
					transOutIn.add(t);
			}
			if (!transOutIn.isEmpty()) {
				Set<PetrinetNode> setN = Sets.newHashSet(p);
				setN.addAll(transOutIn);
				contractGivenPSubnet(net, inputs, outputs, setN, Sets.newHashSet(p), Sets.newHashSet(p), "loop", log);
				return true;				
			}
		}
		return false;
	}

	private static class PlaceWithInput implements Comparable<PlaceWithInput> {
		Place p;
		Transition[] input;

		public PlaceWithInput(Place p, Set<Transition> input) {
			this.p = p;
			this.input = input.toArray(new Transition[] {});

		}

		public int compareTo(PlaceWithInput other) {
			if (input.length < other.input.length)
				return -1;
			else if (input.length > other.input.length)
				return 1;
			else {
				for (int i = 0; i < input.length; i++)
					if (input[i].compareTo(other.input[i]) != 0)
						return input[i].compareTo(other.input[i]);
			}
			return 0;
		}
	}

	private static class TransitionWithInput implements Comparable<TransitionWithInput> {
		Transition t;
		Place[] input;

		public TransitionWithInput(Transition t, Set<Place> input) {
			this.t = t;
			this.input = input.toArray(new Place[] {});

		}

		public int compareTo(TransitionWithInput other) {
			if (input.length < other.input.length)
				return -1;
			else if (input.length > other.input.length)
				return 1;
			else {
				for (int i = 0; i < input.length; i++)
					if (input[i].compareTo(other.input[i]) != 0)
						return input[i].compareTo(other.input[i]);
			}
			return 0;
		}
	}

	public final boolean joinMultiinterface(Petrinet net, Set<PetrinetNode> inputs, Set<PetrinetNode> outputs, ContractionLog log) {		
		//System.err.println("trying to contract nodes with the same interface");
		log.println("trying to contract nodes with the same interface");
		boolean wasSomethingContracted = false;

		//first do for places
		Place[] places = net.getPlaces().toArray(new Place[] {});
		PlaceWithInput[] inputsForPlaces = new PlaceWithInput[places.length];
		for (int i = 0; i < places.length; i++) {
			inputsForPlaces[i] = new PlaceWithInput(places[i], getInTransitions(net, places[i]));
		}
		java.util.Arrays.sort(inputsForPlaces);

		//merge places with the same interface
		int k = 0;
		//System.err.println("miejsca k=0");
		while (k < places.length - 1) {
			//System.err.println("while1 miejsca k="+k);
			Set<PetrinetNode> subnet = new HashSet<>();
			Set<PetrinetNode> subnetIn = new HashSet<>();
			Set<PetrinetNode> subnetOut = new HashSet<>();

			//System.err.println("while2 miejsca k="+k);
			//check if the same input and output
			if (inputsForPlaces[k].compareTo(inputsForPlaces[k + 1]) == 0
					&& getOutTransitions(net, inputsForPlaces[k].p)
							.equals(getOutTransitions(net, inputsForPlaces[k + 1].p))) {
				subnet.add(inputsForPlaces[k].p);
				subnetIn.add(inputsForPlaces[k].p);
				subnetOut.add(inputsForPlaces[k].p);

				subnet.add(inputsForPlaces[k + 1].p);
				subnetIn.add(inputsForPlaces[k + 1].p);
				subnetOut.add(inputsForPlaces[k + 1].p);

				contractGivenPSubnet(net, inputs, outputs, subnet, subnetIn, subnetOut, "multiIO", log);
				wasSomethingContracted = true;
			}
			k++;
		}
		//System.err.println("while koniec miejsca k="+k);

		//next do for transitions
		Transition[] transitions = net.getTransitions().toArray(new Transition[] {});
		TransitionWithInput[] inputsForTransitions = new TransitionWithInput[transitions.length];
		for (int i = 0; i < transitions.length; i++) {
			inputsForTransitions[i] = new TransitionWithInput(transitions[i], getInPlaces(net, transitions[i]));
		}
		java.util.Arrays.sort(inputsForTransitions);

		k = 0;
		//System.err.println("tranzycje k=0");
		//merge transitions with the same interface
		while (k < transitions.length - 1) {
			Set<PetrinetNode> subnet = new HashSet<>();
			Set<PetrinetNode> subnetIn = new HashSet<>();
			Set<PetrinetNode> subnetOut = new HashSet<>();

			//check if the same input and output
			if (inputsForTransitions[k].compareTo(inputsForTransitions[k + 1]) == 0
					&& getOutPlaces(net, inputsForTransitions[k].t)
							.equals(getOutPlaces(net, inputsForTransitions[k + 1].t))) {
				subnet.add(inputsForTransitions[k].t);
				subnetIn.add(inputsForTransitions[k].t);
				subnetOut.add(inputsForTransitions[k].t);

				subnet.add(inputsForTransitions[k + 1].t);
				subnetIn.add(inputsForTransitions[k + 1].t);
				subnetOut.add(inputsForTransitions[k + 1].t);

				contractGivenTSubnet(net, inputs, outputs, subnet, subnetIn, subnetOut, "multiIO", log);
				wasSomethingContracted = true;
			}
			k++;
		}
		//System.err.println("while koniec tranzycje k="+k);

		return wasSomethingContracted;
	}
}
