package pl.mimuw.wfnetreduction.algorithms;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;

//Remembers its function and all elements seen so far. Will return only elements it has not seen before.
class UniqueMapper<E> {
	Set<E> ignoredElements = new HashSet<E>();
	Function<E, E> mapper;
	Predicate<E> predicate;

	private UniqueMapper(Function<E, E> mapper) {
		this.mapper = mapper;
		this.predicate = (E x) -> true;
	}
	
	//factory method
	public static <E> UniqueMapper<E> from(Function<E, E> mapper) {
		return new UniqueMapper<>(mapper);
	}
	
	//add additional predicate
	public void addPredicate(Predicate<E> p) {
		predicate = predicate.and(p);
	}

	E processElement(E e) {
		E dummy = mapper.apply(e);
		if (ignoredElements.add(dummy) && predicate.test(dummy))
			return dummy;
		else
			return null;
	}
}

//A lazy Iterator. Any moment if can be decorated with a map operation and/or filter operation.
//The functions for decoration will be stored in a list and used to process/filter elements.
//At each step duplicates will be eliminated. Storing the functions in a list has this advantage
//over functional implementation that we will not get stack overflow errors.
public class LazyIterator<E> implements Iterator<E> {
	Iterator<E> backingIter;
	E next;

	ArrayList<UniqueMapper<E>> uniqueMappers = new ArrayList<UniqueMapper<E>>();

	private LazyIterator(Iterator<E> backingIter) {
		this.backingIter = backingIter;
		uniqueMappers.add(UniqueMapper.from((E x) -> x));
	}

	//factory method
	public static <T> LazyIterator<T> from(Iterator<T> backingIter) {
		return new LazyIterator<T>(backingIter);
	}

	public void map(Function<E, E> f) {
		uniqueMappers.add(UniqueMapper.from(f));
	}
	
	//filtering is always done after mapping, but we always have the initial dummy/identity mapper
	public void filter(Predicate<E> p) {
		uniqueMappers.get(uniqueMappers.size()-1).addPredicate(p);
	}

	// sets next
	public boolean hasNext() {
		if (next != null)
			return true;

		labelWhile: while (backingIter.hasNext()) {
			next = backingIter.next();

			for (int j = 0; j < uniqueMappers.size(); j++) {
				next = uniqueMappers.get(j).processElement(next);
				if (next == null)
					continue labelWhile;
			}
			if (next != null) 
				return true;
		}

		return false;
	}

	public E next() {
		if (next == null && !hasNext())
			throw new NoSuchElementException();

		E dummy = next;
		next = null;
		return dummy;
	}
}
