package pl.mimuw.wfnetreduction.algorithms;

import java.io.Serializable;
import java.util.Collection;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;

public class PetrinetHelper {
	//clone Petrinet with inputs and outputs
	public static Triple<Petrinet, Set<PetrinetNode>, Set<PetrinetNode>> clonePetrinet(Petrinet net,
			Set<PetrinetNode> inputs, Set<PetrinetNode> outputs) {
		HashMap<DirectedGraphElement, DirectedGraphElement> map = new HashMap<>();
		Petrinet newNet = PetrinetFactory.clonePetrinet(net, map);
		Set<PetrinetNode> newInputs = inputs.stream().map((DirectedGraphElement el) -> (PetrinetNode) map.get(el))
				.collect(Collectors.toSet());
		Set<PetrinetNode> newOutputs = outputs.stream().map((DirectedGraphElement el) -> (PetrinetNode) map.get(el))
				.collect(Collectors.toSet());
		return Triple.of(newNet, newInputs, newOutputs);
	}

	//get in places for a transition
	public static final Set<Place> getInPlaces(Petrinet net, Transition t) {
		HashSet<Place> hs = new HashSet<>();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e1 : net.getInEdges(t)) {
			hs.add((Place) e1.getSource());
		}
		return hs;
	}

	//get in transitions for a place
	public static final Set<Transition> getInTransitions(Petrinet net, Place p) {
		Set<Transition> hs = new HashSet<>();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e1 : net.getInEdges(p)) {
			hs.add((Transition) e1.getSource());
		}
		return hs;
	}

	//get in nodes for a node
	public static final Set<PetrinetNode> getInNodes(Petrinet net, PetrinetNode n) {
		Set<PetrinetNode> hs = new HashSet<>();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e1 : net.getInEdges(n)) {
			hs.add(e1.getSource());
		}
		return hs;
	}

	//get out places for a transition
	public static final Set<Place> getOutPlaces(Petrinet net, Transition t) {
		Set<Place> hs = new HashSet<>();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e1 : net.getOutEdges(t)) {
			hs.add((Place) e1.getTarget());
		}
		return hs;
	}

	//get out transitions for a place
	public static final Set<Transition> getOutTransitions(Petrinet net, Place p) {
		Set<Transition> hs = new HashSet<>();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e1 : net.getOutEdges(p)) {
			hs.add((Transition) e1.getTarget());
		}
		return hs;
	}

	//get out nodes for a node
	public static final Set<PetrinetNode> getOutNodes(Petrinet net, PetrinetNode p) {
		Set<PetrinetNode> hs = new HashSet<>();
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e1 : net.getOutEdges(p)) {
			hs.add(e1.getTarget());
		}
		return hs;
	}

	//get in nodes for a set of nodes
	public static final Set<PetrinetNode> getImmediatePredecessors(Petrinet net, Set<PetrinetNode> start) {
		Set<PetrinetNode> hs = new HashSet<PetrinetNode>();
		for (PetrinetNode n : start) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : net.getInEdges(n)) {
				hs.add(e.getSource());
			}
		}
		return hs;
	}

	//get out nodes for a set of nodes
	public static final Set<PetrinetNode> getImmediateSuccessors(Petrinet net, Set<PetrinetNode> start) {
		Set<PetrinetNode> hs = new HashSet<PetrinetNode>();
		for (PetrinetNode n : start) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : net.getOutEdges(n)) {
				hs.add(e.getTarget());
			}
		}
		return hs;
	}

	public static final Set<Place> getSuccSuccPlaces(Petrinet net, Place start) {
		Set<Transition> trans = getOutTransitions(net, start);
		Set<Place> places = new HashSet<>();
		for (Transition t : trans) {
			places.addAll(getOutPlaces(net, t));
		}
		return places;
	}

	public static final Set<Transition> getSuccSuccTransitions(Petrinet net, Transition start) {
		Set<Place> places = getOutPlaces(net, start);
		Set<Transition> trans = new HashSet<>();
		for (Place p : places) {
			trans.addAll(getOutTransitions(net, p));
		}
		return trans;
	}

	//does not return the origin place
	private static class DownstreamPlaceIterator implements Iterator<Place>, Iterable<Place> {
		Petrinet net;
		Set<Place> visited = new HashSet<>();
		Set<Place> placesToGo = new LinkedHashSet<>();

		public DownstreamPlaceIterator(Petrinet net, Place p) {
			this.net = net;
			placesToGo.addAll(getSuccSuccPlaces(net, p));
			placesToGo.remove(p);
			visited.add(p);
		}

		public boolean hasNext() {
			return !placesToGo.isEmpty();
		}

		public Place next() {
			Place p = placesToGo.iterator().next();
			placesToGo.remove(p);
			visited.add(p);
			Set<Place> succ = getSuccSuccPlaces(net, p);
			succ.removeAll(visited);
			placesToGo.addAll(succ);

			return p;
		}

		public Iterator<Place> iterator() {
			return this;
		}
	}

	public static final Iterator<Place> iterateDownstreamPlaces(Petrinet net, Place start) {
		return new DownstreamPlaceIterator(net, start);
	}

	public static final Iterable<Place> iterableDownstreamPlaces(Petrinet net, Place start) {
		return new DownstreamPlaceIterator(net, start);
	}

	public static final LazyIterator<Pair<Place, Place>> iteratePairsOfDownstreamPlacesNaive(Petrinet net) {
		final Collection<Place> places = net.getPlaces();
		Stream<Pair<Place, Place>> joinedStream = places.stream().flatMap(p -> {
			Iterable<Place> iterable = () -> iterateDownstreamPlaces(net, p);
			return StreamSupport.stream(iterable.spliterator(), false).map(pp -> Pair.of(p, pp));
		});
		return LazyIterator.from(joinedStream.iterator());
	}

	//we only compare the first and second element ignoring the third
	private static class ModifiedTriple<L, M, R> implements Comparable<ModifiedTriple<L, M, R>>, Serializable {

		/** Serialization version */
		private static final long serialVersionUID = 1L;

		/** Left object */
		public final L left;
		/** Middle object */
		public final M middle;
		/** Right object */
		public final R right;

		/**
		 * <p>
		 * Obtains an immutable triple of from three objects inferring the generic
		 * types.
		 * </p>
		 * 
		 * <p>
		 * This factory allows the triple to be created using inference to obtain the
		 * generic types.
		 * </p>
		 * 
		 * @param <L>
		 *            the left element type
		 * @param <M>
		 *            the middle element type
		 * @param <R>
		 *            the right element type
		 * @param left
		 *            the left element, may be null
		 * @param middle
		 *            the middle element, may be null
		 * @param right
		 *            the right element, may be null
		 * @return a triple formed from the three parameters, not null
		 */
		public static <L, M, R> ModifiedTriple<L, M, R> of(final L left, final M middle, final R right) {
			return new ModifiedTriple<L, M, R>(left, middle, right);
		}

		/**
		 * Create a new triple instance.
		 *
		 * @param left
		 *            the left value, may be null
		 * @param middle
		 *            the middle value, may be null
		 * @param right
		 *            the right value, may be null
		 */
		public ModifiedTriple(final L left, final M middle, final R right) {
			super();
			this.left = left;
			this.middle = middle;
			this.right = right;
		}

		//-----------------------------------------------------------------------
		public L getLeft() {
			return left;
		}

		public M getMiddle() {
			return middle;
		}

		public R getRight() {
			return right;
		}

		/**
		 * <p>
		 * Compares the triple based on the left element, followed by the middle
		 * element, finally the right element. The types must be {@code Comparable}.
		 * </p>
		 *
		 * @param other
		 *            the other triple, not null
		 * @return negative if this is less, zero if equal, positive if greater
		 */
		@Override
		public int compareTo(final ModifiedTriple<L, M, R> other) {
			return new CompareToBuilder().append(getLeft(), other.getLeft()).append(getMiddle(), other.getMiddle())
					.toComparison();
		}

		/**
		 * <p>
		 * Compares this triple to another based on the three elements.
		 * </p>
		 *
		 * @param obj
		 *            the object to compare to, null returns false
		 * @return true if the elements of the triple are equal
		 */
		@SuppressWarnings("deprecation") // ObjectUtils.equals(Object, Object) has been deprecated in 3.2
		@Override
		public boolean equals(final Object obj) {
			if (obj == this) {
				return true;
			}
			if (obj instanceof ModifiedTriple<?, ?, ?>) {
				final ModifiedTriple<?, ?, ?> other = (ModifiedTriple<?, ?, ?>) obj;
				return ObjectUtils.equals(getLeft(), other.getLeft())
						&& ObjectUtils.equals(getMiddle(), other.getMiddle());
			}
			return false;
		}

		/**
		 * <p>
		 * Returns a suitable hash code.
		 * </p>
		 *
		 * @return the hash code
		 */
		@Override
		public int hashCode() {
			return (getLeft() == null ? 0 : getLeft().hashCode()) ^ (getMiddle() == null ? 0 : getMiddle().hashCode());
		}

		/**
		 * <p>
		 * Returns a String representation of this triple using the format
		 * {@code ($left,$middle,$right)}.
		 * </p>
		 *
		 * @return a string describing this object, not null
		 */
		@Override
		public String toString() {
			return new StringBuilder().append('(').append(getLeft()).append(',').append(getMiddle()).append(',')
					.append(getRight()).append(')').toString();
		}
	}

	public static final Iterator<Pair<Place, Place>> iteratePairsOfDownstreamPlacesTransitiveClosure(Petrinet net) {
		//we don't consider pairs with the same element, to avoid infinite loops
		Map<Place, Set<Place>> groupedEdges = net.getPlaces().stream().map(p -> {
			Set<Place> s = getSuccSuccPlaces(net, p);
			s.remove(p);
			return Pair.of(p, s);
		}).collect(Collectors.toMap(Pair::getLeft, Pair::getRight));
		//the third element is a distance
		Set<ModifiedTriple<Place, Place, Integer>> edges = groupedEdges.entrySet().stream()
				.flatMap(entry -> entry.getValue().stream().map(p -> ModifiedTriple.of(entry.getKey(), p, 1)))
				.collect(Collectors.toSet());
		Set<ModifiedTriple<Place, Place, Integer>> paths = new HashSet<>();
		Set<ModifiedTriple<Place, Place, Integer>> newPaths = new HashSet<>(edges);
		while (!newPaths.isEmpty()) {
			paths.addAll(newPaths);
			newPaths = newPaths.stream()
					.flatMap(path -> groupedEdges.get(path.getMiddle()).stream().filter(p -> !p.equals(path.getLeft()))
							.map(p -> ModifiedTriple.of(path.getLeft(), p, path.getRight() + 1)))
					.collect(Collectors.toSet());
			newPaths.removeAll(paths);
		}

		return paths.stream().sorted(Comparator.comparing(ModifiedTriple::getRight))
				.map(t -> Pair.of(t.getLeft(), t.getMiddle())).iterator();
	}

	//does not return the origin transition
	private static class DownstreamTransitionIterator implements Iterator<Transition>, Iterable<Transition> {
		Petrinet net;
		Set<Transition> visited = new LinkedHashSet<>();
		Set<Transition> transToGo = new LinkedHashSet<>();

		public DownstreamTransitionIterator(Petrinet net, Transition t) {
			this.net = net;
			transToGo.addAll(getSuccSuccTransitions(net, t));
			transToGo.remove(t);
			visited.add(t);
		}

		public boolean hasNext() {
			return !transToGo.isEmpty();
		}

		public Transition next() {
			Transition t = transToGo.iterator().next();
			transToGo.remove(t);
			visited.add(t);
			Set<Transition> succ = getSuccSuccTransitions(net, t);
			succ.removeAll(visited);
			transToGo.addAll(succ);

			return t;
		}

		public Iterator<Transition> iterator() {
			return this;
		}
	}

	public static final Iterator<Transition> iterateDownstreamTransitions(Petrinet net, Transition start) {
		return new DownstreamTransitionIterator(net, start);
	}

	public static final Iterable<Transition> iterableDownstreamTransitions(Petrinet net, Transition start) {
		return new DownstreamTransitionIterator(net, start);
	}

	public static final LazyIterator<Pair<Transition, Transition>> iteratePairsOfDownstreamTransitionsNaive(Petrinet net) {
		final Collection<Transition> trans = net.getTransitions();
		Stream<Pair<Transition, Transition>> joinedStream = trans.stream().flatMap(t -> {
			Iterable<Transition> iterable = () -> iterateDownstreamTransitions(net, t);
			return StreamSupport.stream(iterable.spliterator(), false).map(tt -> Pair.of(t, tt));
		});
		return LazyIterator.from(joinedStream.iterator());
	}

	public static final Iterator<Pair<Transition, Transition>> iteratePairsOfDownstreamTransitionsTransitiveClosure(
			Petrinet net) {
		//map from transition to set of its immediate successor transitions, but not including the transition itself (to avoid loops)
		Map<Transition, Set<Transition>> succSuccTransitions = net.getTransitions().stream().map(t -> {
			Set<Transition> s = getSuccSuccTransitions(net, t);
			s.remove(t);
			return Pair.of(t, s);
		}).collect(Collectors.toMap(Pair::getLeft, Pair::getRight));

		//covert map to set of tuples with starting trans, successor trans, and distance =1
		Set<ModifiedTriple<Transition, Transition, Integer>> edges = succSuccTransitions.entrySet().stream()
				.flatMap(entry -> entry.getValue().stream().map(t -> ModifiedTriple.of(entry.getKey(), t, 1)))
				.collect(Collectors.toSet());

		//do transitive closure
		Set<ModifiedTriple<Transition, Transition, Integer>> paths = new HashSet<>();
		Set<ModifiedTriple<Transition, Transition, Integer>> newPaths = new HashSet<>(edges);
		//extend one step at a time
		while (!newPaths.isEmpty()) {
			paths.addAll(newPaths);

			newPaths = newPaths.stream()
					.flatMap(path -> succSuccTransitions.get(path.getMiddle()).stream()
							.filter(t -> !t.equals(path.getLeft()))
							.map(t -> ModifiedTriple.of(path.getLeft(), t, path.getRight() + 1)))
					.collect(Collectors.toSet());
			newPaths.removeAll(paths);
		}

		//remove length
		return paths.stream().sorted(Comparator.comparing(ModifiedTriple::getRight))
				.map(t -> Pair.of(t.getLeft(), t.getMiddle())).iterator();
	}

	//check if subnet is cyclic using Kahn algorithm
	public static final boolean checkIfCyclic(Petrinet net, Set<PetrinetNode> subnet) {
		LinkedList<Pair<PetrinetNode, Set<PetrinetNode>>> list = new LinkedList<>();
		for (PetrinetNode n : subnet) {
			Set<PetrinetNode> succ = PetrinetHelper.getOutNodes(net, n);
			succ.retainAll(subnet);
			list.add(Pair.of(n, succ));
		}

		//try to remove nodes without outgoing edges step by step (many nodes can be removed per step)
		while (!list.isEmpty()) {
			//find nodes without outgoing edges
			Set<PetrinetNode> toRemove = new HashSet<>();
			Iterator<Pair<PetrinetNode, Set<PetrinetNode>>> iter = list.iterator();
			while (iter.hasNext()) {
				Pair<PetrinetNode, Set<PetrinetNode>> pair = iter.next();
				if (pair.getRight().isEmpty()) {
					toRemove.add(pair.getLeft());
					iter.remove();
				}
			}

			//if no such nodes, then there is a cycle as this list was not empty
			if (toRemove.isEmpty())
				return true;

			//remove the nodes without outgoing edges
			for (Pair<PetrinetNode, Set<PetrinetNode>> pair : list) {
				pair.getRight().removeAll(toRemove);
			}
		}

		//we removed all nodes so there was no cycles
		return false;
	}

	private enum ANDORClasses {
		_11pOR, _pAND, _11tAND, _tOR
	};

	public static final Triple<Set<PetrinetNode>, Set<PetrinetNode>, Set<PetrinetNode>> expandP(Petrinet net,
			Set<PetrinetNode> in, Set<PetrinetNode> out, Pair<Place, Place> startPair, ContractionLog log,
			boolean limitTo11) {
		log.println("trying to expand from " + startPair);
		Place i = startPair.getLeft();
		Place o = startPair.getRight();
		Set<PetrinetNode> subnet = new HashSet<>();
		subnet.add(i);
		subnet.add(o);

		Set<PetrinetNode> subIn = new HashSet<>();
		subIn.add(i);
		boolean hasGlobalIn = in.contains(i);
		boolean hasNonGlobalIn = !hasGlobalIn;

		Set<PetrinetNode> subOut = new HashSet<>();
		subOut.add(o);
		boolean hasGlobalOut = out.contains(o);
		boolean hasNonGlobalOut = !hasGlobalOut;

		Set<PetrinetNode> tba = new HashSet<>();
		tba.addAll(getOutNodes(net, i));
		tba.addAll(getInNodes(net, o));
		subnet.addAll(tba);

		//this is method version for i,o being places
		EnumSet<ANDORClasses> pc = EnumSet.of(ANDORClasses._11pOR, ANDORClasses._pAND);

		Set<PetrinetNode> presetForInputs = getInNodes(net, i);
		Set<PetrinetNode> postsetForOutputs = getOutNodes(net, o);

		while (!tba.isEmpty() && !pc.isEmpty()) {
			Iterator<PetrinetNode> dummy = tba.iterator();
			PetrinetNode n = dummy.next();
			dummy.remove();

			Set<PetrinetNode> presetForN = getInNodes(net, n);
			Set<PetrinetNode> postsetForN = getOutNodes(net, n);

			if (presetForN.equals(presetForInputs) && (in.contains(n) ? hasGlobalIn : hasNonGlobalIn)) {
				pc.remove(ANDORClasses._11pOR);
				//pc.remove(ANDORClasses._11tAND);
				if (pc.isEmpty())//no possible net classes
					return null;

				subIn.add((Place) n);
				if (limitTo11)
					return null;

			} else {
				//non global input cannot be non interface node
				if (in.contains(n))
					return null;
				Set<PetrinetNode> dummy2 = new HashSet<>(presetForN);
				dummy2.removeAll(subnet);
				tba.addAll(dummy2);
				subnet.addAll(dummy2);
			}

			if (postsetForN.equals(postsetForOutputs) && (out.contains(n) ? hasGlobalOut : hasNonGlobalOut)) {
				pc.remove(ANDORClasses._11pOR);
				//pc.remove(ANDORClasses._11tAND);
				if (pc.isEmpty())//no possible net classes
					return null;

				subOut.add((Place) n);
				
				if (limitTo11)
					return null;

			} else {
				//non global input cannot be non interface node
				if (out.contains(n))
					return null;
				Set<PetrinetNode> dummy2 = new HashSet<>(postsetForN);
				dummy2.removeAll(subnet);
				tba.addAll(dummy2);
				subnet.addAll(dummy2);
			}

			presetForN.retainAll(subnet);
			postsetForN.retainAll(subnet);
			//check if multiple edges for single node are allowed
			if ((!subIn.contains(n) && presetForN.size() != 1) || (subIn.contains(n) && presetForN.size() != 0)
					|| (!subOut.contains(n) && postsetForN.size() != 1)
					|| (subOut.contains(n) && postsetForN.size() != 0)) {
				if (n instanceof Place) {
					pc.remove(ANDORClasses._pAND);
					//pc.remove(ANDORClasses._11tAND);
					if (pc.isEmpty())
						return null;
				} else if (n instanceof Transition) {
					//pc.remove(ANDORClasses._tOR);
					pc.remove(ANDORClasses._11pOR);
					if (pc.isEmpty())
						return null;
				} else {
					throw new RuntimeException("The node is neither place nor transition.");
				}
			}
		}

		if (checkIfCyclic(net, subnet)) {
			pc.remove(ANDORClasses._pAND);
			//pc.remove(ANDORClasses._11tAND);
		}

		if (pc.isEmpty())
			return null;

		//if (subIn.stream().allMatch(t -> in.contains(t)) || subIn.stream().allMatch(t -> !in.contains(t)))
		return Triple.of(subnet, subIn, subOut);
	}

	public static final Triple<Set<PetrinetNode>, Set<PetrinetNode>, Set<PetrinetNode>> expandT(Petrinet net,
			Set<PetrinetNode> in, Set<PetrinetNode> out, Pair<Transition, Transition> startPair, ContractionLog log, boolean limitTo11) {
		log.println("trying to expand from " + startPair);
		Transition i = startPair.getLeft();
		Transition o = startPair.getRight();
		Set<PetrinetNode> subnet = new HashSet<>();
		subnet.add(i);
		subnet.add(o);

		Set<PetrinetNode> subIn = new HashSet<>();
		subIn.add(i);
		boolean hasGlobalIn = in.contains(i);
		boolean hasNonGlobalIn = !hasGlobalIn;

		Set<PetrinetNode> subOut = new HashSet<>();
		subOut.add(o);
		boolean hasGlobalOut = out.contains(o);
		boolean hasNonGlobalOut = !hasGlobalOut;

		Set<PetrinetNode> tba = new HashSet<>();
		tba.addAll(getOutNodes(net, i));
		tba.addAll(getInNodes(net, o));
		subnet.addAll(tba);

		//this is method version for i,o being transitions
		EnumSet<ANDORClasses> pc = EnumSet.of(ANDORClasses._11tAND, ANDORClasses._tOR);

		Set<PetrinetNode> presetForInputs = getInNodes(net, i);
		Set<PetrinetNode> postsetForOutputs = getOutNodes(net, o);

		while (!tba.isEmpty() && !pc.isEmpty()) {
			Iterator<PetrinetNode> dummy = tba.iterator();
			PetrinetNode n = dummy.next();
			dummy.remove();
			Set<PetrinetNode> presetForN = getInNodes(net, n);
			Set<PetrinetNode> postsetForN = getOutNodes(net, n);

			//by the argument in the proof this is enough if we don't consider the global interface nodes (or even simpler without checking nonGlobal)
			//if (hasGlobalIn ? in.contains(n) : (!in.contains(n) && presetForN.equals(presetForInputs))) {
			//the stronger version is necessary because the global net does not have to be and-or, so the argument from the proof does not have to hold
			if (presetForN.equals(presetForInputs) && (in.contains(n) ? hasGlobalIn : hasNonGlobalIn)) {
				//pc.remove(ANDORClasses._11pOR);
				pc.remove(ANDORClasses._11tAND);
				if (pc.isEmpty())//no possible net classes
					return null;
				//pc.remove(ANDORClasses._11tAND);

				subIn.add((Transition) n);
				
				if (limitTo11)
					return null;

			} else {
				//non global input cannot be non interface node
				if (in.contains(n))
					return null;
				Set<PetrinetNode> dummy2 = new HashSet<>(presetForN);
				dummy2.removeAll(subnet);
				tba.addAll(dummy2);
				subnet.addAll(dummy2);
			}

			if (postsetForN.equals(postsetForOutputs) && (out.contains(n) ? hasGlobalOut : hasNonGlobalOut)) {
				//pc.remove(ANDORClasses._11pOR);
				pc.remove(ANDORClasses._11tAND);
				if (pc.isEmpty())//no possible net classes
					return null;

				if (n.getClass() == Place.class) {
					System.err.println("test, coś poszło nie tak");
					System.err.println(n.toString());
					System.err.println(i.toString());
					System.err.println(o.toString());
				}
				//subOut.add((Transition) n);
				subOut.add(n);
				
				if (limitTo11)
					return null;
			} else {
				//non global input cannot be non interface node
				if (out.contains(n))
					return null;
				Set<PetrinetNode> dummy2 = new HashSet<>(postsetForN);
				dummy2.removeAll(subnet);
				tba.addAll(dummy2);
				subnet.addAll(dummy2);
			}

			presetForN.retainAll(subnet);
			postsetForN.retainAll(subnet);
			if ((!subIn.contains(n) && presetForN.size() != 1) || (subIn.contains(n) && presetForN.size() != 0)
					|| (!subOut.contains(n) && postsetForN.size() != 1)
					|| (subOut.contains(n) && postsetForN.size() != 0)) {
				if (n instanceof Place) {
					//pc.remove(ANDORClasses._pAND);
					pc.remove(ANDORClasses._11tAND);
					if (pc.isEmpty())
						return null;
				} else if (n instanceof Transition) {
					pc.remove(ANDORClasses._tOR);
					//pc.remove(ANDORClasses._11pOR);
					if (pc.isEmpty())
						return null;
				} else {
					throw new RuntimeException("The node is neither place nor transition.");
				}
			}
		}

		if (checkIfCyclic(net, subnet)) {
			//pc.remove(ANDORClasses._pAND);
			pc.remove(ANDORClasses._11tAND);
		}

		if (pc.isEmpty())
			return null;
		//if (subIn.stream().allMatch(t -> in.contains(t)) || subIn.stream().allMatch(t -> !in.contains(t)))
		return Triple.of(subnet, subIn, subOut);
	}

	//prints out edges of the PN
	static public String toString(Petrinet pn) {
		return pn.getEdges().stream().map(edge -> edge.getSource().getLabel() + "->" + edge.getTarget())
				.collect(Collectors.joining(", "));
	}

	//	//prints out input, edges, output
	//	static public String toString(Triple<Petrinet, Set<Place>, Set<Place>> triple) {
	//		return "I=" + triple.getMiddle().toString() + " " + toString(triple.getLeft()) + " O="+triple.getRight().toString();
	//	}
	//
	//prints out input, edges, output
	static public String toString(Triple<Petrinet, Set<PetrinetNode>, Set<PetrinetNode>> triple) {
		return "I=" + triple.getMiddle().toString() + " " + toString(triple.getLeft()) + " O="
				+ triple.getRight().toString();
	}

	private static final boolean doesTPairContainAnyOfTheNodesInSet(Pair<Transition, Transition> p,
			Set<PetrinetNode> set) {
		return set.contains(p.getLeft()) || set.contains(p.getRight());
	}

	private static final boolean doesPPairContainAnyOfTheNodesInSet(Pair<Place, Place> p, Set<PetrinetNode> set) {
		return set.contains(p.getLeft()) || set.contains(p.getRight());
	}

	private static final boolean doesTPairContainBothNodesFromTheSets(Pair<Transition, Transition> p,
			Set<PetrinetNode> setStart, Set<PetrinetNode> setEnd) {
		return setStart.contains(p.getLeft()) && setEnd.contains(p.getRight());
	}

	private static final boolean doesPPairContainBothNodesFromTheSets(Pair<Place, Place> p, Set<PetrinetNode> setStart,
			Set<PetrinetNode> setEnd) {
		return setStart.contains(p.getLeft()) && setEnd.contains(p.getRight());
	}

	private static final <NODE extends PetrinetNode> Pair<NODE, NODE> replaceWithNodeIntoWhichWeContractIfNeeded(NODE n,
			Pair<NODE, NODE> p, Set<PetrinetNode> subnet) {
		return subnet.contains(p.getLeft()) ? Pair.of(n, p.getRight())
				: subnet.contains(p.getRight()) ? Pair.of(p.getLeft(), n) : p;
	}

	//filters out pairs of downstream transition and place pairs assuming that we contract into place
	static public void filterDownstreamStreamsP(LazyIterator<Pair<Place, Place>> iteratorPP, LazyIterator<Pair<Transition, Transition>> iteratorTT,
			Set<PetrinetNode> netM, Set<PetrinetNode> inputsM, Set<PetrinetNode> outputsM, Place p) {
		Set<PetrinetNode> interiorNodes = new HashSet<PetrinetNode>(netM);
		interiorNodes.removeAll(inputsM);
		interiorNodes.removeAll(outputsM);
		iteratorTT.filter(pair -> !doesTPairContainAnyOfTheNodesInSet(pair, interiorNodes));
		iteratorPP.filter(pair -> !doesPPairContainAnyOfTheNodesInSet(pair, interiorNodes));
		iteratorPP.filter(pair -> !doesPPairContainBothNodesFromTheSets(pair, inputsM, outputsM));
		iteratorPP.map(pair -> replaceWithNodeIntoWhichWeContractIfNeeded(p, pair, netM));
	}

	//filters out pairs of downstream transition and place pairs assuming that we contract into transition
	static public void filterDownstreamStreamsT(LazyIterator<Pair<Place, Place>> iteratorPP, LazyIterator<Pair<Transition, Transition>> iteratorTT,
			Set<PetrinetNode> netM, Set<PetrinetNode> inputsM, Set<PetrinetNode> outputsM, Transition t) {
		Set<PetrinetNode> interiorNodes = new HashSet<PetrinetNode>(netM);
		interiorNodes.removeAll(inputsM);
		interiorNodes.removeAll(outputsM);
		iteratorTT.filter(pair -> !doesTPairContainAnyOfTheNodesInSet(pair, interiorNodes));
		iteratorTT.filter(pair -> !doesTPairContainBothNodesFromTheSets(pair, inputsM, outputsM));
		iteratorTT.map(pair -> replaceWithNodeIntoWhichWeContractIfNeeded(t, pair, netM));
		iteratorPP.filter(pair -> !doesPPairContainAnyOfTheNodesInSet(pair, interiorNodes));
	}
}